package editor;

import java.io.File;
import java.io.IOException;

import editor.model.DocumentView;
import editor.model.FileHandler;
import editor.model.WindowView;
import javafx.fxml.FXML;
import javafx.scene.control.ChoiceBox;
import javafx.scene.control.ScrollBar;
import javafx.scene.control.TextField;
import javafx.scene.input.KeyCode;
import javafx.scene.layout.Pane;
import javafx.scene.shape.Line;
import javafx.scene.text.Font;
import javafx.scene.text.Text;
import javafx.stage.FileChooser;
import javafx.stage.Stage;

public class Controller {
	private double charWidth, charHeight;
	private WindowView visibleText;
	private final double HEIGHT_FUDGE = 1;
	private Stage mainStage;
	
	@FXML
	private Pane textPane;
	@FXML
	private ScrollBar scroller;
	@FXML
	private TextField msg;
	@FXML
	private ChoiceBox<String> choices;
	private ClassFinder<DocumentView> finder = new ClassFinder<>(DocumentView.class, "editor.model.implementations");
	
	private Text content;
	private Font textFont;
	private Line cursor;
	
	private FileChooser chooser = new FileChooser();
	
	int getRowCapacity() {
		return (int)Math.ceil(textPane.heightProperty().get() / charHeight);
	}
	
	int getCharsPerLine() {
		return (int)(textPane.widthProperty().get() / charWidth);
	}
	
	@FXML
	private void initialize() {
		textPane.setOnKeyPressed(k -> arrows(k.getCode()));
		textPane.setOnKeyTyped(k -> type(k.getCharacter()));
		
		content = new Text();
		textFont = new Font("Monaco", 12);
		content.setFont(textFont);
		content.setText("A");
		charWidth = content.getLayoutBounds().getWidth();
		charHeight = content.getLayoutBounds().getHeight() + HEIGHT_FUDGE;
		content.setText("");
		
		cursor = new Line();
		
		textPane.getChildren().add(content);
		textPane.getChildren().add(cursor);
		content.setX(0);
		content.setY(12);
		
		finder.loadChoiceNames(choices);
	}
	
	private void orientCursor() {
		double x = visibleText.getCursorColumn() * charWidth;
		double y = visibleText.getCursorRow() * charHeight;
		cursor.setStartX(x);
		cursor.setEndX(x);
		cursor.setStartY(y);
		cursor.setEndY(y + charHeight);
	}
	
	private void ensureText() throws InstantiationException, IllegalAccessException {
		if (visibleText == null) {
			visibleText = new WindowView(finder.newInstanceOf(choices.getSelectionModel().getSelectedItem()), getRowCapacity(), getCharsPerLine());
		}		
	}
	
	@FXML
	private void focus() {
		try {
			textPane.requestFocus();
			ensureText();
		} catch (Exception exc) {
			msg.setText(exc.getMessage());
		}
	}
	
	@FXML
	private void display() { 
		content.setText(visibleText.getVisibleText());
		orientCursor();
		msg.setText("");
	}

	private void type(String chrString) {
		if (chrString.length() == 1) {
			visibleText.keystroke(chrString.charAt(0));
			display();
		}
	}
	
	private void arrows(KeyCode k) {
		if (k == KeyCode.LEFT) {
			visibleText.goLeft();
			display();
		} else if (k == KeyCode.RIGHT) {
			visibleText.goRight();
			display();
		} else if (k == KeyCode.BACK_SPACE) {
			visibleText.backspace();
			display();
		} else if (k == KeyCode.UP) {
			visibleText.goUp();
			display();
		} else if (k == KeyCode.DOWN) {
			visibleText.goDown();
			display();
		}
	}
	
	@FXML
	void open() {
		chooser.setTitle("Open text file");
		File selectedFile = chooser.showOpenDialog(mainStage);
		if (selectedFile != null) {
			try {
				ensureText();
				FileHandler.openFrom(selectedFile, visibleText);
				display();
			} catch (Exception e) {
				msg.setText(e.getMessage());
			}
		}
	}
	
	@FXML
	void save() {
		chooser.setTitle("Save text file");
		File selectedFile = chooser.showSaveDialog(mainStage);
		if (selectedFile != null) {
			try {
				FileHandler.saveTo(selectedFile, visibleText);
			} catch (IOException e) {
				msg.setText(e.getMessage());
			}
		}
	}
	
	@FXML
	void newDoc() {
		try {
			visibleText = null;
			ensureText();
			display();
		} catch (Exception exc) {
			exc.printStackTrace();
			msg.setText(exc.getMessage());
		}
	}
	
	void setStage(Stage mainStage) {
		this.mainStage = mainStage;
	}
}
