package editor.model;

public class WindowView {
	private int cursorRow, cursorCol, totalRows, totalCols;
	private DocumentView doc;
	
	public WindowView(DocumentView doc, int totalRows, int totalCols) {
		this(doc, 0, 0, totalRows, totalCols);
	}
	
	public WindowView(DocumentView doc, int cursorRow, int cursorCol, int totalRows, int totalCols) {
		this.doc = doc;
		this.cursorRow = cursorRow;
		this.cursorCol = cursorCol;
		this.totalRows = totalRows;
		this.totalCols = totalCols;
	}
	
	public int getNumRows() {return totalRows;}
	public int getNumCols() {return totalCols;}
	
	public int getCursorRow() {return cursorRow;}
	public int getCursorColumn() {return cursorCol;}
	
	public String getVisibleText() {
		StringBuilder result = new StringBuilder();
		int charCount = 0;
		int lineCount = 0;
		DocumentView cursor = getWindowStart();
		while (lineCount < getNumRows() && !cursor.atEnd()) {
			char c = cursor.peekAfter();
			result.append(c);
			if (c == '\n') {
				lineCount++;
				charCount = 0;
			} else {
				charCount++;
				if (charCount % totalCols == 0) {
					lineCount++;
					result.append('\n');
				}
			}
			cursor.goRight();
		}
		return result.toString();
	}
	
	public String getCompleteText() {
		return doc.getAllChars();
	}
	
	public void replaceTextWith(String src) {
		doc.replaceWith(src);
		while (!doc.atStart()) {
			doc.goLeft();
		}
		cursorCol = cursorRow = 0;
	}
	
	DocumentView getWindowStart() {
		int charCount = 0;
		int lineCount = 0;
		DocumentView cursor = doc.shallowCopy();
		while (!cursor.atStart()) {
			char c = cursor.peekBefore();
			if (c == '\n') {
				lineCount++;
				charCount = 0;
			} else {
				charCount++;
				if (charCount % totalCols == 0) {
					lineCount++;
				}
			}
			if (lineCount > cursorRow) {
				return cursor;
			} else {
				cursor.goLeft();
			}
		}
		return cursor;
	}
	
	public void keystroke(char key) {
		if (key == '\r') {key = '\n';}
		doc.insert(key);
		advanceColumn(key);
	}
	
	public boolean atStart() {return doc.atStart();}
	public boolean atEnd() {return doc.atEnd();}
	public char peekBefore() {return doc.peekBefore();}
	public char peekAfter() {return doc.peekAfter();}
	
	public void goLeft() {
		if (!atStart()) {
			doc.goLeft();
			cursorBack();
		}
	}
	
	public void goRight() {
		if (!atEnd()) {
			advanceColumn(doc.peekAfter());
			doc.goRight();
		}
	}
	
	public void backspace() {
		if (!atStart()) {
			doc.removeBefore();
			cursorBack();
		}
	}
	
	public void goUp() {
		int targetCol = cursorCol;
		while (cursorCol > 0) goLeft();
		goLeft();
		while (cursorCol > targetCol) goLeft();
	}

	public void goDown() {
		int targetCol = cursorCol;
		if (cursorCol == 0) {
			goRight();
			if (atEnd() || peekAfter() == '\n') {
				return;
			}
		} 
		while (cursorCol != 0 && !atEnd()) goRight();
		while (cursorCol < targetCol && !atEnd()) goRight();
	}

	private void cursorBack() {
		cursorCol -= 1;
		if (cursorCol < 0) {
			cursorCol = doc.getPreviousLine().length() % totalCols;
			if (cursorRow > 0) {
				cursorRow -= 1;
			}
		}
	}

	private void advanceColumn(char key) {
		if (key == '\n') {
			cursorCol = 0;
			if (cursorRow < totalRows - 1) {
				cursorRow += 1;
			}
		} else {
			cursorCol += 1;
			if (cursorCol >= totalCols) {
				cursorCol = 0;
				if (cursorRow < totalRows - 1) {
					cursorRow += 1;
				}
			}
		}
	}
}
