package editor.model;

import static org.junit.Assert.*;

import org.junit.Test;

import editor.model.implementations.DoubleTextNodeView;

public class WindowViewTest {
	final static String bigRow1 = "This is a longer test string that we are using.\n";
	final static String bigRow2 = "It has four rows and can spill over several columns very easily.\n";
	final static String bigRow3 = "What will the consequences be?\n";
	final static String bigRow4 = "This is the fourth and final line of text we will employ.";
	final static String bigText = bigRow1 + bigRow2 + bigRow3 + bigRow4;
	
	final static String replacement = "This is the replacement text.";
	
	WindowView win = new WindowView(new DoubleTextNodeView(), 5, 35);
	
	String windowEndText;
	
	String row2split(String row, int numCols) {
		StringBuilder result = new StringBuilder();
		int split = 0;
		while (split + numCols < row.length()) {
			result.append(row.substring(split, split + numCols));
			result.append('\n');
			split += numCols;
		}
		result.append(row.substring(split));
		return result.toString();
	}
	
	@Test
	public void visibleTextTest0() {
		for (int i = 0; i < bigText.length(); i++) {
			win.keystroke(bigText.charAt(i));
		}
		assertEquals(4, win.getCursorRow());
		windowEndText = row2split(bigRow2, win.getNumCols()) + 
				row2split(bigRow3, win.getNumCols()) + 
				row2split(bigRow4, win.getNumCols());
		assertEquals(windowEndText, win.getVisibleText());
	}
	
	@Test
	public void completeTextTest() {
		visibleTextTest0();
		assertEquals(bigText, win.getCompleteText());
	}
	
	@Test
	public void replaceTest() {
		completeTextTest();
		win.replaceTextWith(replacement);
		assertEquals(replacement, win.getCompleteText());
		assertEquals(replacement, win.getVisibleText());
		assertTrue(win.atStart());
	}
	
	@Test
	public void backspaceTest() {
		visibleTextTest0();
		genericLeftTest(win::backspace);
		assertTrue(win.atEnd());
		assertTrue(win.atStart());
	}
	
	@Test	
	public void visibleTextTest1() {
		visibleTextTest0();
		for (int i = 0; i < win.getNumCols(); i++) {
			win.goLeft();
		}
		assertEquals(3, win.getCursorRow());
		assertEquals(windowEndText, win.getVisibleText());
		
		for (int i = 0; i < win.getNumCols() * 3; i++) {
			win.goLeft();
		}
		assertEquals(0, win.getCursorRow());
		assertEquals(windowEndText, win.getVisibleText());	
	}
	
	@Test
	public void visibleTextTest2() {
		visibleTextTest1();
		
		while (win.peekAfter() != '\n') {
			win.goLeft();
		}
		assertEquals(0, win.getCursorRow());
		assertEquals(bigRow1.substring(bigRow1.length() - win.getNumCols()) +
					row2split(bigRow2, win.getNumCols()) + 
					row2split(bigRow3, win.getNumCols()) + 
					bigRow4.substring(0, win.getNumCols()) + "\n",
				win.getVisibleText());		
	}
	
	@Test
	public void visibleTextTest3() {
		visibleTextTest2();
		
		while (!win.atStart()) {
			win.goLeft();
		}
		
		assertEquals(row2split(bigRow1, win.getNumCols()) + 
					row2split(bigRow2, win.getNumCols()) +
					bigRow3,
				win.getVisibleText());
	}
	
	@Test
	public void keystrokeTest() {
		win.keystroke('a');
		assertEquals('a', win.peekBefore());
		win.keystroke('b');
		assertEquals('b', win.peekBefore());
		win.keystroke('\r');
		assertEquals('\n', win.peekBefore());
		assertEquals("ab\n", win.getVisibleText());
	}
	
	@Test
	public void afterTest() {
		keystrokeTest();
		win.goLeft();
		assertEquals('\n', win.peekAfter());
		win.goLeft();
		assertEquals('b', win.peekAfter());
		win.goLeft();
		assertEquals('a', win.peekAfter());
	}
	
	@Test
	public void goRightTest() {
		visibleTextTest3();
		int wasCol = win.getCursorColumn();
		int wasRow = win.getCursorRow();
		while (!win.atEnd()) {
			win.goRight();
			if (win.getCursorColumn() == 0) {
				if (wasRow + 1 < win.getNumRows()) {
					assertEquals(wasRow + 1, win.getCursorRow());
				}
			} else {
				assertEquals(wasCol + 1, win.getCursorColumn());
			}
			wasRow = win.getCursorRow();
			wasCol = win.getCursorColumn();
		}
		
		win.goRight();
		assertTrue(win.atEnd());
	}
	
	@Test
	public void goLeftTest() {
		goRightTest();
		genericLeftTest(win::goLeft);
	}
	
	private void genericLeftTest(final Runnable lefter) {
		int wasCol = win.getCursorColumn();
		int wasRow = win.getCursorRow();
		while (!win.atStart()) {
			lefter.run();
			if (win.getCursorColumn() > wasCol) {
				if (wasRow > 0) {
					assertEquals(wasRow - 1, win.getCursorRow());
				}
			} else {
				assertEquals(wasCol - 1, win.getCursorColumn());
			}
			wasCol = win.getCursorColumn();
			wasRow = win.getCursorRow();
		}
		
		lefter.run();
		assertTrue(win.atStart());
	}
	
	@Test
	public void goUpTest() {
		visibleTextTest0();
		win.goUp();
		assertEquals('d', win.peekBefore());
		assertEquals(' ', win.peekAfter());
		win.goUp();
		assertEquals('e', win.peekBefore());
		assertEquals('n', win.peekAfter());
		win.goUp();
		assertEquals(' ', win.peekBefore());
		assertEquals('e', win.peekAfter());
		win.goUp();
		assertEquals('c', win.peekBefore());
		assertEquals('a', win.peekAfter());
		win.goUp();
		assertFalse(win.atStart());
		assertEquals('.', win.peekBefore());
		assertEquals('\n', win.peekAfter());
		win.goUp();
		assertEquals('o', win.peekBefore());
		assertEquals('n', win.peekAfter());
		win.goUp();
		assertTrue(win.atStart());
		win.goUp();
		assertTrue(win.atStart());
	}
	
	@Test
	public void goDownTest() {
		goUpTest();
		for (int i = 0; i < 12; i++) {
			win.goRight();
		}
		assertEquals('o', win.peekBefore());
		assertEquals('n', win.peekAfter());
		win.goDown();
		assertEquals('.', win.peekBefore());
		assertEquals('\n', win.peekAfter());
		win.goDown();
		assertEquals(' ', win.peekBefore());
		assertEquals('r', win.peekAfter());
		win.goDown();
		assertEquals('l', win.peekBefore());
		assertEquals('u', win.peekAfter());
		win.goDown();
		assertEquals('h', win.peekBefore());
		assertEquals('e', win.peekAfter());
		win.goDown();
		assertEquals(' ', win.peekBefore());
		assertEquals('f', win.peekAfter());
		win.goDown();
		assertEquals('i', win.peekBefore());
		assertEquals('l', win.peekAfter());
		win.goDown();
		assertTrue(win.atEnd());
		win.goDown();
		assertTrue(win.atEnd());
	}
	
	@Test
	public void goDownTest2() {
		goUpTest();
		assertTrue(win.atStart());
		win.goDown();
		assertFalse(win.atStart());
		System.out.println(win.getVisibleText());
		System.out.println(win.peekBefore() + ", " + win.peekAfter());
		assertEquals('w', win.peekBefore());
		assertEquals('e', win.peekAfter());
	}
	
	@Test
	public void allNewlines() {
		win.keystroke('\n');
		win.keystroke('\n');
		assertTrue(win.atEnd());
		win.goDown();
		assertTrue(win.atEnd());
		win.goUp();
		win.goUp();
		assertTrue(win.atStart());
		win.goDown();
		assertEquals('\n', win.peekBefore());
		assertEquals('\n', win.peekAfter());
		win.goDown();
		assertTrue(win.atEnd());
	}
	
	@Test
	public void goDownLeft() {
		win.replaceTextWith("This is a test.\nOver.");
		for (int i = 0; i < 8; i++) {
			win.goRight();
		}
		win.goDown();
		assertTrue(win.atEnd());
	}
}
