package editor.model.implementations;

import static org.junit.Assert.*;

import org.junit.Test;

public class DoubleTextNodeTest {
	
	DoubleTextNode node1 = new DoubleTextNode('a');
	DoubleTextNode node2 = new DoubleTextNode('b');
	DoubleTextNode node3 = new DoubleTextNode('c');
	DoubleTextNode node4 = new DoubleTextNode('d');
	
	public void checkAddAfter(DoubleTextNode node, DoubleTextNode added) {
		DoubleTextNode old = new DoubleTextNode(node);
		node.addAfter(added);
		assertTrue(node.isConsistent());
		assertTrue(added.isConsistent());
		assertTrue(node.getNext() == added);
		assertTrue(added.getPrev() == node);
		assertTrue(added.getNext() == old.getNext());
		assertTrue(added.getNext() == null || old.getNext().getPrev() == added);
	}
	
	public void checkAddBefore(DoubleTextNode node, DoubleTextNode added) {
		DoubleTextNode old = new DoubleTextNode(node);
		node.addBefore(added);
		assertTrue(node.isConsistent());
		assertTrue(added.isConsistent());
		assertTrue(node.getPrev() == added);
		assertTrue(added.getNext() == node);
		assertTrue(added.getPrev() == old.getPrev());
		assertTrue(added.getPrev() == null || old.getPrev().getNext() == added);
	}
	
	public void checkRemoveBefore(DoubleTextNode node) {
		DoubleTextNode old = new DoubleTextNode(node);
		node.removeBefore();
		assertTrue(node.isConsistent());
		if (old.getPrev() == null) {
			assertEquals(old, node);
		} else {
			assertTrue(node.getPrev() == old.getPrev().getPrev());
			if (node.getPrev() != null) {
				assertTrue(node.getPrev().getNext() == node);
			}
		}
	}
	
	public void checkRemoveAfter(DoubleTextNode node) {
		DoubleTextNode old = new DoubleTextNode(node);
		node.removeAfter();
		assertTrue(node.isConsistent());
		if (old.getNext() == null) {
			assertEquals(old, node);
		} else {
			assertTrue(node.getNext() == old.getNext().getNext());
			if (node.getNext() != null) {
				assertTrue(node.getNext().getPrev() == node);
			}
		}
	}
	
	public void checkMergeBefore(DoubleTextNode node, DoubleTextNode other) {
		DoubleTextNode old = new DoubleTextNode(node);
		if (other == null) {
			node.mergeBefore(other);
			assertEquals(old, node);
		} else {
			DoubleTextNode oldOther = new DoubleTextNode(other);
			node.mergeBefore(other);
			assertTrue(node.isConsistent());
			assertTrue(other.isConsistent());
			assertTrue(node.getPrev() == other);
			assertTrue(other.getNext() == node);
			assertTrue(node.getNext() == old.getNext());
			assertTrue(other.getPrev() == oldOther.getPrev());
		}
	}

	public void checkMergeAfter(DoubleTextNode node, DoubleTextNode other) {
		DoubleTextNode old = new DoubleTextNode(node);
		if (other == null) {
			node.mergeAfter(other);
			assertEquals(old, node);
		} else {
			DoubleTextNode oldOther = new DoubleTextNode(other);
			node.mergeAfter(other);
			assertTrue(node.isConsistent());
			assertTrue(other.isConsistent());
			assertTrue(node.getNext() == other);
			assertTrue(other.getPrev() == node);
			assertTrue(node.getPrev() == old.getPrev());
			assertTrue(other.getNext() == oldOther.getNext());
		}
	}
	
	public void assertListsEqual(DoubleTextNode i1, DoubleTextNode i2) {
		while (i1 != null && i2 != null) {
			assertEquals(i1.getChar(), i2.getChar());
			i1 = i1.getNext();
			i2 = i2.getNext();
		}
		assertTrue(i1 == null && i2 == null);
	}
	
	public void checkCopyBefore(DoubleTextNode node) {
		DoubleTextNode copy = node.copyBefore();
		assertTrue(node.isConsistent());
		assertTrue(copy.isConsistent());
		assertEquals(null, copy.getNext());
		while (node.getPrev() != null) {
			assertEquals(node.getChar(), copy.getChar());
			node = node.getPrev();
			copy = copy.getPrev();
		}
	}
	
	public void checkCopyAfter(DoubleTextNode node) {
		DoubleTextNode copy = node.copyAfter();
		assertTrue(node.isConsistent());
		assertTrue(copy.isConsistent());
		assertEquals(null, copy.getPrev());
		while (node.getNext() != null) {
			assertEquals(node.getChar(), copy.getChar());
			node = node.getNext();
			copy = copy.getNext();
		}
	}
	
	@Test
	public void testAddAfter1() {
		checkAddAfter(node1, node2);
		checkAddAfter(node2, node3);
	}
	
	@Test
	public void testAddBefore1() {
		checkAddBefore(node1, node2);
		checkAddBefore(node2, node3);
	}
	
	@Test
	public void testAddAfter2() {
		testAddAfter1();
		checkAddAfter(node2, node4);
	}
	
	@Test
	public void testAddBefore2() {
		testAddBefore1();
		checkAddBefore(node2, node4);
	}
	
	@Test
	public void testRemoveAfter() {
		testAddBefore1();
		checkRemoveAfter(node3);
		checkRemoveAfter(node3);
		checkRemoveAfter(node3);
	}
	
	@Test
	public void testRemoveBefore() {
		testAddAfter1();
		checkRemoveBefore(node3);
		checkRemoveBefore(node3);
		checkRemoveBefore(node3);
	}
	
	@Test
	public void testCopyBefore() {
		testAddAfter2();
		checkCopyBefore(node3);
	}
	
	@Test
	public void testCopyAfter() {
		testAddBefore2();
		checkCopyAfter(node3);
	}
	
	@Test
	public void testMergeBefore() {
		checkAddAfter(node1, node2);
		checkAddAfter(node3, node4);
		checkMergeBefore(node3, node2);
		checkMergeBefore(node1, null);
	}
	
	@Test
	public void testMergeAfter() {
		checkAddAfter(node1, node2);
		checkAddAfter(node3, node4);
		checkMergeAfter(node2, node3);
		checkMergeAfter(node4, null);
	}
	
	@Test
	public void testChars() {
		assertEquals('a', node1.getChar());
		assertEquals('b', node2.getChar());
		assertEquals('c', node3.getChar());
		assertEquals('d', node4.getChar());
	}
}
